package nl.wur.bif.pantools.analysis.export_pangenome;

import jakarta.validation.constraints.Pattern;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.BeanUtils;
import picocli.CommandLine.Model.CommandSpec;

import java.nio.file.Path;
import java.util.concurrent.Callable;

import static picocli.CommandLine.*;

/**
 * Convert a pangenome graph to a different format.
 * @author Dirk-Jan van Workum, Wageningen University, the Netherlands.
 */
@Command(name = "export_gfa", sortOptions = false)
public class ExportGfaCLI implements Callable<Integer> {
    @Spec CommandSpec spec;

    @ParentCommand
    private Pantools pantools;

    @Option(names = {"-o", "--output"}, required = true)
    private Path outputPath;

    @Option(names = "--format")
    @Pattern(regexp = "GFAv1|GFAv2", message = "{pattern.graph-format}")
    private String outputFormat;

    @Override
    public Integer call() throws Exception {
        pantools.createLogger(spec);
        BeanUtils.argValidation(spec, this);

        pantools.setPangenomeGraph("pangenome");

        PangenomeExporterGfa pangenomeExporterGfa = new PangenomeExporterGfa(outputPath);
        pangenomeExporterGfa.convert(outputFormat);

        return 0;
    }
}
