package nl.wur.bif.pantools.analysis.export_pangenome;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.graph.IndexGraphExtractor;
import nl.wur.bif.pantools.utils.graph.graph_objects.Graph;
import nl.wur.bif.pantools.utils.writer.Gfa1Writer;
import nl.wur.bif.pantools.utils.writer.Gfa2Writer;
import nl.wur.bif.pantools.utils.writer.GfaWriter;

import java.io.IOException;
import java.nio.file.Path;

/**
 * Convert a pangenome graph to a different format.
 * NB: Currently only GFA is supported.
 * @author Dirk-Jan van Workum, Wageningen University, the Netherlands.
 */
public class PangenomeExporterGfa {
    private final Path outputPath;

    /**
     * Constructor.
     * NB: Currently only GFA is supported.
     * @param outputPath The path to the output file.
     */
    public PangenomeExporterGfa(Path outputPath) {
        this.outputPath = outputPath;
    }

    /**
     * Convert the pangenome graph to a different format.
     * @param format The format to convert to.
     * @throws IOException If an IO error occurs during conversion.
     */
    public void convert(String format) throws IOException {
        Pantools.logger.debug("Converting pangenome");

        // start graph extractor
        IndexGraphExtractor indexGraphExtractor = new IndexGraphExtractor();
        Graph graph = indexGraphExtractor.extractNucleotideGraph();

        // write the graph
        GfaWriter gfaWriter;
        switch (format) {
            case "GFAv1":
                gfaWriter = new Gfa1Writer();
                break;
            case "GFAv2":
                gfaWriter = new Gfa2Writer();
                break;
            default:
                Pantools.logger.error("Unsupported format: {}", format);
                throw new IllegalArgumentException("Unsupported format");
        }
        Pantools.logger.trace("Graph: {}", graph);
        gfaWriter.write(graph, outputPath);
    }
}
