package nl.wur.bif.pantools.utils.graph.graph_objects;

import java.util.Map;
import java.util.Set;

public class Graph {
    final private Map<Long, GraphNode> graphNodes;
    final private Map<Long, GraphEdge> graphEdges;
    final private Set<GraphPath> graphPaths; //walks in the graph as determined by "origin" (e.g.: G2S3 (genome 2, sequence 3))

    public Graph(Map<Long, GraphNode> graphNodes, Map<Long, GraphEdge> graphEdges, Set<GraphPath> graphPaths) {
        this.graphNodes = graphNodes;
        this.graphEdges = graphEdges;
        this.graphPaths = graphPaths;
    }

    /**
     * Walks all paths in the graph and checks if they are valid
     * @return true if all paths are valid, false otherwise
     */
    public boolean isValid() {
        if (graphPaths == null) {
            return false;
        }
        for (GraphPath graphPath : graphPaths) {
            if (!graphPath.isValid(this)) {
                return false;
            }
        }
        return true;
    }

    public Map<Long, GraphNode> getGraphNodes() {
        return graphNodes;
    }

    public Map<Long, GraphEdge> getEdges() {
        return graphEdges;
    }

    public Set<GraphPath> getGraphPaths() {
        return graphPaths;
    }

    public boolean contains(GraphNode graphNode) {
        if (graphNodes != null) {
            return graphNodes.containsValue(graphNode); //this is slow, I know
        } else {
            return false;
        }
    }

    public boolean contains(GraphEdge graphEdge) {
        if (graphEdges != null) {
            return graphEdges.containsValue(graphEdge); //this is slow, I know
        } else {
            return false;
        }
    }

    public boolean contains(GraphPath graphPath) {
        if (graphPaths != null) {
            return graphPaths.contains(graphPath);
        } else {
            return false;
        }
    }

    @Override
    public String toString() {
        return "Graph{" +
                "graphNodes=" + graphNodes +
                ", graphEdges=" + graphEdges +
                ", graphPaths=" + graphPaths +
                '}';
    }
}
