package nl.wur.bif.pantools.utils.graph.graph_objects;

import htsjdk.samtools.Cigar;
import nl.wur.bif.pantools.utils.Globals;
import org.neo4j.graphdb.RelationshipType;

import java.util.Arrays;
import java.util.Set;
import java.util.HashMap;

public class GraphEdge {
    final private long id;
    final private GraphNode source;
    final private char sourceOrientation;
    final private GraphNode target;
    final private char targetOrientation;
    final private RelationshipType type;
    final private Cigar cigar;

    public GraphEdge(long id, GraphNode source, GraphNode target, RelationshipType type) {
        this.id = id;
        this.source = source;
        this.target = target;
        this.type = type;
        this.cigar = Cigar.fromCigarString(String.format("%dM", Globals.K_SIZE - 1)); //TODO: because cDBG; will need to be generalized for other graphs

        switch (type.name()) {
            case "FF":
                this.sourceOrientation = '+';
                this.targetOrientation = '+';
                break;
            case "FR":
                this.sourceOrientation = '+';
                this.targetOrientation = '-';
                break;
            case "RF":
                this.sourceOrientation = '-';
                this.targetOrientation = '+';
                break;
            case "RR":
                this.sourceOrientation = '-';
                this.targetOrientation = '-';
                break;
            default:
                throw new RuntimeException(String.format("Unknown edge type %s", type));
        }
    }

    public long getId() {
        return id;
    }

    public GraphNode getSource() {
        return source;
    }

    public char getSourceOrientation() {
        return sourceOrientation;
    }

    public GraphNode getTarget() {
        return target;
    }

    public char getTargetOrientation() {
        return targetOrientation;
    }

    public RelationshipType getType() {
        return type;
    }

    public Cigar getCigar() {
        return cigar;
    }

    public boolean equals(GraphEdge graphEdge) {
        return this.id == graphEdge.getId();
    }

    public String toString() {
        return String.format("Edge %d: %s -> %s", id, source, target);
    }
}
