package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

public class FastaWriter {
    public FastaWriter() {
    }

    /**
     * Writes all the sequences in the graph range to a separate fasta file per sequence (AKA origin).
     * @param graphRange The graph range to write.
     * @param outputDir The output directory.
     * @throws IOException If an error occurs while creating the directory of the fasta files.
     */
    public void write(GraphRange graphRange, Path outputDir) throws IOException {
        Pantools.logger.info("Writing the graph as fasta files to {}/", outputDir);

        // create the directory if it does not exist
        if (!Files.exists(outputDir)) {
            Files.createDirectories(outputDir);
        }

        // Go over the genome ranges in the graph range and add to the appropriate fasta file
        graphRange.getGenomeRanges().forEach(genomeRange -> {
            final Path fastaFile = outputDir.resolve(genomeRange.getGenomeNr() + ".fasta");
            final String fastaHeader = genomeRange.getCleanedFastaHeader();
            Pantools.logger.debug("Writing sequence '{}' for sequence '{}'.", fastaHeader, fastaFile);

            // Append the sequences to the fasta file
            try (FileWriter fw = new FileWriter(fastaFile.toFile(), true);
                 BufferedWriter writer = new BufferedWriter(fw)) {
                writer.write(fastaHeader);
                writer.newLine();
                writer.write(genomeRange.getNucleotideSequence());
                writer.newLine();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        });
    }
}
