package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.utils.graph.graph_objects.GraphEdge;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphPath;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphNode;

public class Gfa1Writer extends GfaWriter {
    private static final String GFA_VERSION = "1.0";
    private static final char GFA_PATH_SEPARATOR = ',';

    public Gfa1Writer() {
        super();
    }

    @Override
    protected String getGraphNodeOutputString(GraphNode graphNode) {
        //Column 	Field 	Type 	Regexp 	Description
        //1 	RecordType 	Character 	S 	Record type
        //2 	Name 	String 	[!-)+-<>-~][!-~]* 	Segment name
        //3 	Sequence 	String 	\*\|[A-Za-z=.]+ 	Optional nucleotide sequence
        //4 	OtherFields 	String 	.* 	Optional fields
        return String.format("S\t%d\t%s\tDP:f:%d\n",
                graphNode.getId(),
                graphNode.getSequence(),
                graphNode.getFrequency()
        );
    }

    @Override
    protected String getEdgeOutputString(GraphEdge graphEdge) {
        //Column 	Field 	Type 	Regexp 	Description
        //1 	RecordType 	Character 	L 	Record type
        //2 	From 	String 	[!-)+-<>-~][!-~]* 	Name of segment
        //3 	FromOrient 	String 	+\|- 	Orientation of From segment
        //4 	To 	String 	[!-)+-<>-~][!-~]* 	Name of segment
        //5 	ToOrient 	String 	+\|- 	Orientation of To segment
        //6 	Overlap 	String 	\*\|([0-9]+[MIDNSHPX=])+ 	Optional CIGAR string describing overlap
        return String.format("L\t%d\t%c\t%d\t%c\t%s\n",
                graphEdge.getSource().getId(),
                graphEdge.getSourceOrientation(),
                graphEdge.getTarget().getId(),
                graphEdge.getTargetOrientation(),
                graphEdge.getCigar().toString()
        );
    }

    @Override
    protected String getGraphPathOutputString(GraphPath graphPath) {
        //1 	RecordType 	Character 	P 	Record type
        //2 	PathName 	String 	[!-)+-<>-~][!-~]* 	Path name
        //3 	SegmentNames 	String 	[!-)+-<>-~][!-~]* 	A comma-separated list of segment names and orientations
        //4 	Overlaps 	String 	\*\|([0-9]+[MIDNSHPX=])+ 	Optional comma-separated list of CIGAR strings
        return String.format("P\t%s\t%s\t*\n",
                graphPath.getOrigin(),
                convertGraphPathToGfa(graphPath)
        );
    }

    @Override
    protected String getGfaVersion() {
        return GFA_VERSION;
    }

    @Override
    protected char getGraphPathSeparator() {
        return GFA_PATH_SEPARATOR;
    }
}
