package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.utils.graph.graph_objects.GraphPath;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphEdge;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphNode;

public class Gfa2Writer extends GfaWriter {
    private static final String GFA_VERSION = "2.0";
    private static final char GFA_PATH_SEPARATOR = ' ';

    public Gfa2Writer() {
        super();
    }

    @Override
    protected String getGraphNodeOutputString(GraphNode graphNode) {
        //S <sid:id> <slen:int> <sequence> <tag>*
        return String.format("S\t%d\t%d\t%s\n",
                graphNode.getId(),
                graphNode.getLength(),
                graphNode.getSequence()
        );
    }

    @Override
    protected String getEdgeOutputString(GraphEdge graphEdge) {
        //E <eid:opt_id> <sid1:ref> <sid2:ref> <beg1:pos> <end1:pos> <beg2:pos> <end2:pos> <alignment> <tag>*
        return String.format("E\t%d\t%s%c\t%s%s\t%d\t%d$\t%d\t%d\t%s\n",
                graphEdge.getId(),
                graphEdge.getSource().getId(),
                graphEdge.getSourceOrientation(),
                graphEdge.getTarget().getId(),
                graphEdge.getTargetOrientation(),
                graphEdge.getSource().getLength() - graphEdge.getCigar().getReferenceLength(),
                graphEdge.getSource().getLength(), //the $ denoting the end of the overlap is hardcoded
                0,
                graphEdge.getCigar().getReadLength(),
                graphEdge.getCigar().toString()
        );
    }

    @Override
    protected String getGraphPathOutputString(GraphPath graphPath) {
        //O <oid:opt_id> <ref>([ ]<ref>)* <tag>*
        return String.format("O\t%s\t%s\t*\n",
                graphPath.getOrigin(),
                convertGraphPathToGfa(graphPath)
        );
    }

    @Override
    protected String getGfaVersion() {
        return GFA_VERSION;
    }

    @Override
    protected char getGraphPathSeparator() {
        return GFA_PATH_SEPARATOR;
    }
}
