package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.graph.GenomeRange;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import org.neo4j.graphdb.*;

import java.io.IOException;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;

public class HomologyPresenceAbsenceWriter extends PresenceAbsenceWriter {
    @Override
    public void write(GraphRange graphRange, Path outputFilePath) throws IOException {
        Pantools.logger.info("Writing homology group presence/absence to {}", outputFilePath);

        // obtain all homology groups that have mRNAs in the given graphRange and determine presence/absence in other genomes
        Map<Long, Map<Integer, Long>> homologyGroupPresenceAbsence = new HashMap<>(); // homology group -> genome -> copy number for hm in said genome
        Map<Long, String> homologyGroupMetadata = new HashMap<>(); // homology group -> concatenated list of mRNA IDs (separated by semicolon)
        try (Transaction ignored = Globals.GRAPH_DB.beginTx();
             ResourceIterator<Node> hmNodes = Globals.GRAPH_DB.findNodes(Globals.HOMOLOGY_GROUP_LABEL)) {
            while (hmNodes.hasNext()) {
                final Node hmNode = hmNodes.next();
                final long hmId = hmNode.getId();
                for (Relationship rel : hmNode.getRelationships(Globals.RelTypes.has_homolog, Direction.OUTGOING)) {
                    final Node mrnaNode = rel.getEndNode();
                    if (!mrnaNode.hasLabel(Globals.MRNA_LABEL)) {
                        continue;
                    }
                    final int[] address = (int[]) mrnaNode.getProperty("address");
                    final int genomeNr = address[0];
                    if (!Globals.selectedGenomes.contains(genomeNr)) {
                        continue;
                    }
                    final String id = String.format("%d#%s", genomeNr, mrnaNode.getProperty("id"));
                    if (graphRange.contains(new GenomeRange(address, false))) {
                        homologyGroupPresenceAbsence.computeIfAbsent(hmId, k -> new HashMap<>()).merge(genomeNr, 1L, Long::sum);
                        homologyGroupMetadata.putIfAbsent(hmId, "");
                        homologyGroupMetadata.computeIfPresent(hmId, (k, v) -> v.isEmpty() ? id : v + ";" + id);
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find nodes in the graph database.");
            throw new RuntimeException(nfe);
        }

        // write presence/absence as TSV file
        super.writeTsv(homologyGroupPresenceAbsence, homologyGroupMetadata, outputFilePath);
    }
}
