package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

public class PhasingWriter {
    public PhasingWriter() {
    }

    /**
     * Writes the phasing information present in the graph range to a TSV file.
     * @param graphRange The graph range to write.
     * @param outputTsv The output TSV file with three columns: the first column contains the sequence name, the second column contains the original sequence name and the third column contains the phasing information.
     * @throws IOException If an error occurs while creating the directory of the fasta files.
     */
    public void write(GraphRange graphRange, Path outputTsv) throws IOException {
        Pantools.logger.info("Writing the phasing information to {}", outputTsv);

        // create the directory of the file if it does not exist
        if (!Files.exists(outputTsv.getParent())) {
            Files.createDirectories(outputTsv.getParent());
        }

        try (BufferedWriter writer = Files.newBufferedWriter(outputTsv)) {
            // write header
            writer.write("sequence_header\toriginal_name\tphasing_ID");
            writer.newLine();

            // write phasing information
            graphRange.getGenomeRanges().forEach(genomeRange -> {
                try {
                    writer.write(genomeRange.getCleanedSequenceHeader());
                    writer.write("\t");
                    writer.write(genomeRange.getSequenceName());
                    writer.write("\t");
                    writer.write(genomeRange.getPhasingInformation());
                    writer.newLine();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            });
        }
    }
}
