package nl.wur.bif.pantools.utils.writer;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.Globals;
import nl.wur.bif.pantools.utils.graph.graph_objects.GraphRange;
import org.neo4j.graphdb.*;

import java.io.IOException;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;

public class UnitigPresenceAbsenceWriter extends PresenceAbsenceWriter {
    @Override
    public void write(GraphRange graphRange, Path outputFilePath) throws IOException {
        Pantools.logger.info("Writing unitig presence/absence to {}", outputFilePath);

        // obtain all nodes in the given graphRange and determine presence/absence in all genomes
        Map<Long, Map<Integer, Long>> nodePresenceAbsence = new HashMap<>(); // node -> genome -> copy number for hm in said genome
        Map<Long, String> nodeMetadata = new HashMap<>(); // node -> nucleotide sequence
        try (Transaction ignored = Globals.GRAPH_DB.beginTx()) {
            for (long nodeId : graphRange.getNucleotideNodes()) {
                final Node node = Globals.GRAPH_DB.getNodeById(nodeId);
                final String sequence = (String) node.getProperty("sequence");
                nodeMetadata.put(nodeId, sequence);
                final long[] frequencies = (long[]) node.getProperty("frequencies");
                for (int genomeNr = 1; genomeNr < frequencies.length; genomeNr++) {
                    if (Globals.selectedGenomes.contains(genomeNr)) {
                        nodePresenceAbsence.computeIfAbsent(nodeId, k -> new HashMap<>()).put(genomeNr, frequencies[genomeNr]);
                    }
                }
            }
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Could not find nodes in the graph database.");
            throw new RuntimeException(nfe);
        }

        // write presence/absence as TSV file
        super.writeTsv(nodePresenceAbsence, nodeMetadata, outputFilePath);
    }
}
